#! /usr/bin/python

"""//Web service to know the benefit of using cloud resource to improve a batch's QoS//

- > The function displays the various makespan information for the batch, to help the user/scheduler to decide to use QoS resources or not. 

- File: calculate_cloud_benefit.py
 
- Arguments - Using the HTTP GET method:
 - batch_id --- The batch identifier, as a string
 - current_time (Optional, for testing) ---  See info.get_elapsed_time
 - use_last_db_entry (Optional, for testing) --- See info.get_elapsed_time
 - > If the batch_id argument is missing, a web form will be displayed.
 
- Displays a Web page containing:
```
<div class="qos">T1</div>
<div class="qos">T2</div>
<div class="qos">T3</div>
<div class="qos">N</div>
```
With: 
 - T1 the expected makespan for batch without tail effect
 - T2 the expected makespan for batch under tail effect
 - T3 the expected makespan for batch using cloud resources
 - N is the cost of using cloud resources (n_jobs/10, will probably change)
    
- Notes: 
 - A database configuration file named "info_db.cfg" must be present in the directory to connect to the SpeQuloS Info module database.

"""

import sys
import os
sys.path.append(os.path.join(sys.path[0],".."))
os.chdir(sys.path[0])
from tools import conf_func
from tools import db_func
from tools import cgi_func
from tools import web_func
from tools import log_func
log=log_func.get_spequlos_log(__file__)

from get_elapsed_time import get_elapsed_time 
from get_percent_completed import get_percent_completed
from get_n_jobs import get_n_jobs

import math

def calculate_cloud_benefit(conn,batch_id,current_time=None,use_last_db_entry=False):
    """Function implementing the calculate_cloud_benefit operation

    Calculates the benefits and the cost of using some cloud ressources for the last 10% of a batch's jobs. 
    
    Arguments:
    conn -- The Info's DB connection, as a MySQL.connections.Connection object
    batch_id -- The batch identifier, as a string
    current_time (optional) -- A fake "current_time", as an integer denoting the number of seconds since 1970-01-01 00:00:00 UTC.
    use_last_db_time (optional) -- Boolean flag to use the timestamp found in the last databese entry related to this batch as a fake current time.  
    
    Returns a tuple in the form:
    (DG makespan (without tail problem: T_90v+T_10v), Expected DG makespan (with tail problem: T_90v*2), 
    Cloud assisted makespan (T_90v+T_10c), QoS cost in credits (n_jobs/10) )
    
    """   
    if not isinstance(conn,db_func.db_connection) or conn.open==0:
        log.error("The conn argument must be an open connection to the database !")
        return None      
    if not isinstance(batch_id,str):
        log.error("calculate_cloud_benefit() batch_id argument must be a string")
        return None
    if current_time is not None and (not isinstance(current_time,int) or current_time<0):
        log.error("calculate_cloud_benefit() current_time argument must be None or a positive integer")
        return None
    if use_last_db_entry is not None and (not isinstance(use_last_db_entry,bool)):
        log.error("calculate_cloud_benefit() use_last_db_entry argument must be None or a boolean")
        return None    

    if use_last_db_entry != False:
        elapsed = get_elapsed_time(conn,batch_id,use_last_db_entry=True)
        percent_compl = get_percent_completed(conn,batch_id)
    elif current_time !=None:
        elapsed = get_elapsed_time(conn,batch_id,current_time)
        percent_compl = get_percent_completed(conn,batch_id,current_time)
    else:
        elapsed = get_elapsed_time(conn,batch_id)
        percent_compl = get_percent_completed(conn,batch_id)

    n_jobs = get_n_jobs(conn,batch_id)
    if (n_jobs == None):
        log.error("n_jobs is None, cannot compute calculate_cloud_benefit")
        return None
   
    T_90v = compute_T_dg(90,elapsed,percent_compl)
    if (T_90v == None):
        log.error("T_90v is None, cannot compute calculate_cloud_benefit")
        return None
    
    T_10v = compute_T_dg(10,elapsed,percent_compl)
    if (T_10v == None):
        log.error("T_10v is None, cannot compute calculate_cloud_benefit")
        return None

    T_10c = compute_T_cloud(10,elapsed,percent_compl)	 
    if (T_10c == None):
        log.error("T_10c is None, cannot compute calculate_cloud_benefit")
        return None

    return ( int(T_90v+T_10v), int(T_90v*2), int(T_90v+T_10c), int(math.ceil(float(n_jobs)/10))) 


def compute_T_dg(x,elapsed_time,percent_completed):
    """//Computes T(x) for DG, for arbitrary x by extrapolating T(percent_completed) == elapsed_time//
        
    - > T(x) denotes the time elapsed since batch submission when x percents of its jobs have been completed. To compute this time for arbitrary X with assume constant rate for jobs completion, i.e.:
    ```
    T(x)/x == T(percent_completed)/percent_completed == elapsed_time/percent_completed
    ```
        
    - Arguments:
     - x -- The percentage of batch completion 	
     - elapsed_time -- The batch's elapsed time since its submission, as an integer
     - percent_completed -- The batch completion ratio, as a float comprised between 0 and 1
        
    - The function returns the number of seconds since jobs submission to compute x percent of the batch's jobs, as an integer
        
    """     
    if not isinstance(x,int) or x<0 or x>100:
        log.error("compute_T_dg() x argument must be an integer between 0 and 100")
        return None
    if not isinstance(elapsed_time,int) or x<0:
        log.error("compute_T_dg() elapsed_time argument must be a positive integer")
        return None
    if not isinstance(percent_completed,float) or percent_completed<=0 or percent_completed>1:
        log.error("compute_T_dg() percentage_completed argument must be a float between 0 and 1")
        return None

    t_x = int(elapsed_time*x/percent_completed)
    return t_x


def compute_T_cloud(x,elapsed_time,percent_completed):
    """//Computes T(x) for cloud// 

    - > As we don't have any information on cloud performance, we currently use the formula:
    ``` 
    T(x) for Cloud == T(x) for DG / 3
    ```     
    - Arguments:
     - x -- The percentage of batch completion     
     - elapsed_time -- The batch's elapsed time since its submission, as an integer
     - percent_completed -- The batch completion ratio, as a float comprised between 0 and 1
 
    - The function returns the number of seconds since jobs submission to compute x percent of the batch's jobs in the cloud, as an integer 
        
    """
    if not isinstance(x,int) or x<0 or x>100:
        log.error("compute_T_dg() x argument must be an integer between 0 and 100")
        return None
    if not isinstance(elapsed_time,int) or x<0:
        log.error("compute_T_dg() elapsed_time argument must be a positive integer")
        return None
    if not isinstance(percent_completed,float) or percent_completed<=0 or percent_completed>1:
        log.error("compute_T_dg() percent_completed argument must be a float between 0 and 1")
        return None

    tdg = int(elapsed_time*x/percent_completed)
    if isinstance(tdg,int) and tdg>=0:
        return int(tdg/3)
    else:
        log.error("Unable to compute T_cloud: T_dg is not a positive integer")
        return None


if __name__=="__main__":
    cgi_func.print_header()

    if cgi_func.get_form_value('batch_id') is None:
        cgi_func.print_form(['batch_id','current_time','use_last_db_entry'])

    else:
        batch_id = cgi_func.get_form_value('batch_id')
            
        if cgi_func.get_form_value('use_last_db_entry') is not None:
            try:
                use_last_db=bool(int(cgi_func.get_form_value('use_last_db_entry')))
            except ValueError as e:
                log.error("use_last_db_entry must be 0 or 1")
                cgi_func.print_form(['batch_id','current_time','use_last_db_entry'])            
            else:
                conn=db_func.connect(db_func.read_db_config("info_db.cfg"))
                res=calculate_cloud_benefit(conn,batch_id,use_last_db_entry=use_last_db)
                db_func.disconnect(conn)

        elif cgi_func.get_form_value('current_time') is not None:
            try:
                cur_time=int(cgi_func.get_form_value('current_time'))
            except ValueError as e:
                log.error("current_time must be an integer")
                cgi_func.print_form(['batch_id','current_time','use_last_db_entry'])            
            else:
                conn=db_func.connect(db_func.read_db_config("info_db.cfg"))
                res=calculate_cloud_benefit(conn,batch_id,current_time=cur_time)
                db_func.disconnect(conn)

        else:
            conn=db_func.connect(db_func.read_db_config("info_db.cfg"))
            res=calculate_cloud_benefit(conn,batch_id)
            db_func.disconnect(conn)

        print "-----------------------------------------<br/>"
        print "BatchId: %s<br/>" % (batch_id)
        print "-----------------------------------------<br/>"
        print "Computing cloud benefits...<br/>"
        for r in res:
            cgi_func.print_result(r)
        print "-----------------------------------------<br/>"

    cgi_func.print_footer()

