#! /usr/bin/python

"""//Web service to know the percentage of completion of a batch//
 
- > The percentage of completion of a batch is the percentage of his jobs already completed compared to its total number of jobs.

- File: get_percentage_completed.py

- Arguments - Using the HTTP GET method:
 - batch_id --- The batch identifier, as a string
 - current_time (Optional, for testing) ---  Retrieves batch percentage completion at specified fake "current_time", as an integer denoting the number of seconds since 1970-01-01 00:00:00 UTC. 
 - > If the batch_id argument is missing, a web form will be displayed.
 
- Displays a Web page containing:
 - ``<div class="qos">F</div>`` with F the completion ratio between 0 and 1     
 
- Notes: 
 - A database configuration file named "info_db.cfg" must be present in the directory to connect to the SpeQuloS Info module database.

"""
import sys
import os
sys.path.append(os.path.join(sys.path[0],".."))
os.chdir(sys.path[0])
from tools import db_func
from tools import cgi_func
from tools import log_func
log=log_func.get_spequlos_log(__file__)


def get_percent_completed(conn,batch_id,current_time=None):
    """Function implementing the get_percentage_completed operation

    Arguments:
    conn -- The Info's DB connection, as a MySQL.connections.Connection object
    batch_id -- The batch identifier, as a string
    current_time (optional) -- A fake "current_time", as an integer denoting the number of seconds since 1970-01-01 00:00:00 UTC.
     
    The function returns the batch's completion ratio, as a float comprised between 0 and 1 

    """
    if not isinstance(conn,db_func.db_connection) or conn.open==0:
        log.error("The conn argument must be an open connection to the database !")
        return None       
    if not isinstance(batch_id,str):
        log.error("The batch_id argument must be a string !")
        return None
    if current_time is not None and (not isinstance(current_time,int) or current_time<=0):
        log.error("The current_time argument must be a positive integer !")
        return None
                       
    c=conn.cursor()
    
    log.info("Computing percentage of completion of batch %s %s" % (batch_id,"" if current_time is None else "stopping at "+str(current_time)))
    if current_time is None:
        try:
            c.execute("SELECT n_jobs_completed,n_jobs_uncompleted FROM BatchCompletionHistory WHERE batch_id='%s' AND ts=\
                (SELECT MAX(ts) FROM BatchCompletionHistory WHERE batch_id='%s')" % (batch_id,batch_id))
        except db_func.DBError as e:
            log.error("Error during the SQL command execution !")
            log.error(e)
            return None              
    else:
        try:
            c.execute("SELECT n_jobs_completed,n_jobs_uncompleted FROM BatchCompletionHistory WHERE batch_id='%s' AND ts=\
                (SELECT MAX(ts) FROM BatchCompletionHistory WHERE batch_id='%s' AND ts<=%d)" % (batch_id,batch_id,current_time))
        except db_func.DBError as e:
            log.error("Error during the SQL command execution !")
            log.error(e)
            return None      
    
    res = c.fetchall()
    
    if len(res)==0:
        log.warn("No entry with such batch identifier in the database : %s" % batch_id )
        return None
    elif len(res)>1:
        log.error("Several results with same timestamp and batch_id %s in the database ! '(ts,batch_id)' should be BatchCompletionHistory PRIMARY KEY ! "% batch_id)
        return None
    
    try:
        job_compl = int(res[0][0])
        if job_compl<0:
            raise ValueError("Negative n_jobs_completed !")
    except ValueError as e:
        log.error("The 'n_jobs_completed' in BatchCompletionHistory table must be a positive or zero integer !")
        log.error(e)
        return None        
    try:
        job_ucompl = int(res[0][1])
        if job_ucompl<0:
            raise ValueError("Negative n_jobs_uncompleted !")
    except ValueError as e:
        log.error("The 'n_jobs_uncompleted' in BatchCompletionHistory table must be a positive or zero integer !")
        log.error(e)
        return None        
        
    p_completed = float(job_compl)/(job_compl+job_ucompl)
    log.info("Completion is %f" % p_completed)
    return p_completed 

if __name__=="__main__":
    cgi_func.print_header()

    if cgi_func.get_form_value('batch_id') is None:
        cgi_func.print_form(['batch_id','current_time'])

    else:
        batch_id = cgi_func.get_form_value('batch_id')
            
        if cgi_func.get_form_value('current_time') is not None:
            try:
                cur_time=int(cgi_func.get_form_value('current_time'))
            except ValueError as e:
                log.error("cur_time argument must be an integer")
                cgi_func.print_form(['batch_id','current_time'])
            else:
                conn=db_func.connect(db_func.read_db_config("info_db.cfg"))
                cgi_func.print_result( get_percent_completed(conn,batch_id,current_time=cur_time))
                db_func.disconnect(conn)

        else:
            conn=db_func.connect(db_func.read_db_config("info_db.cfg"))
            cgi_func.print_result( get_percent_completed(conn,batch_id))
            db_func.disconnect(conn)
        
    cgi_func.print_footer()

