#!/usr/bin/python

"""//Cloud Wroker implementation on Eucalyptus//

- > Contains some function to start and stop Cloud Worker instance hosted on Eucalyptus, using the libcloud API

- File: euca_cloudworkers.py

- Notes: 
 - A file, named "euca.cfg", with Eucalyptus host (EUCA_HOST and EUCA_PORT), credentials (EUCA_ACCESS_KEY and EUCA_SECRET_KEY) and instance information (EUCA_CW_IMAGE_ID and EUCA_CW_SIZE_ID) are needed to start the instances. 
 - A file, named "id_rsa", containing the private key to connect with SSH to the cloud worker instance on Eucalyptus, must be present in the current directory.
 
"""

from libcloud.types import Provider as LC_ProviderType
import libcloud.providers as LC_providers
from libcloud.types import NodeState as LC_NodeState
import paramiko

import time
import threading

import sys
import os
sys.path.append(os.path.join(sys.path[0],"..",".."))
os.chdir(sys.path[0])

from tools import conf_func
from tools import log_func
log=log_func.get_spequlos_log(__file__)


class Euca_CW_Starter(threading.Thread):
    """//Thread that starts a cloud worker in Eucalyptus and make it executes a list of commands//

    - > In the current implmementation, it makes few minutes for an instance to start.
    
    - Arguments:
     - list_of_commands -- The list of commands to be executed by the cloud worker, as a list of stings
     
    - Attributes:
     - execution_is_success -- When the thread terminated, True if the cloud worker correctly started and the execution of the list of commands succeed, False otherwise
     - specific_data -- When the thread terminated and the execution succeed, may contain some specific data related to the batch. For Eucalyptus, it contains the instance identifier. 
    
    """     
    
    def __init__(self,list_of_commands):
        threading.Thread.__init__(self)
        if not isinstance(list_of_commands,list) or len(list_of_commands)==0:
            raise Exception("No command to execute")
        for s in list_of_commands:
            if not isinstance(s,str):
                raise Exception("Invalid list of commands (must be a list of string)")
        
        self.euca_param = conf_func.read_config("euca.cfg")
        if not ("EUCA_ACCESS_KEY" in self.euca_param and \
                "EUCA_SECRET_KEY" in self.euca_param and \
                "EUCA_HOST" in self.euca_param and \
                "EUCA_CW_IMAGE_ID" in self.euca_param and \
                "EUCA_CW_SIZE_ID" in self.euca_param):
            raise Exception("Missing parameters in \"euca.cfg\"")
        
        if not "EUCA_PORT" in self.euca_param:
            self.euca_param["EUCA_PORT"] = 8773 

        self.list_of_commands = list_of_commands
        self.execution_is_success=False
        self.specific_data=None


    def run(self):
        log.info("Connecting to Eucalyptus at %s" % self.euca_param["EUCA_HOST"])
        
        Driver = LC_providers.get_driver(LC_ProviderType.EUCALYPTUS)
        conn = Driver(self.euca_param["EUCA_ACCESS_KEY"], self.euca_param["EUCA_SECRET_KEY"],\
                host=self.euca_param["EUCA_HOST"],port=self.euca_param["EUCA_PORT"]) 
        
        log.info("Retrieving the list of Eucalyptus images")
        try:
            images = conn.list_images() 
        except Exception as e:
            log.error("Error when retrieving the list of images !")
            log.error(e)
            return None
        
        cw_img = None
        for i in images:
            if self.euca_param["EUCA_CW_IMAGE_ID"] == i.id:
                cw_img=i        

        if cw_img == None:
            log.error("Cannot find CloudWorker image in Eucalyptus server")
            return         

        log.info("CloudWorker image found as %s" % str(cw_img))

        log.info("Reading through the list of Eucalyptus instance types to find the type we want for the CW")
        sizes = conn.list_sizes() 
        cw_size = None
        for s in sizes:
            if self.euca_param["EUCA_CW_SIZE_ID"] == s.id:
                cw_size=s
        
        if cw_size == None:
            log.error("Cannot find CloudWorker size in Eucalyptus server")
            return 

        log.info("CloudWorker instance type found as %s" % str(cw_size))
        
        log.info("Starting the instance")
        cw_node_launch = conn.create_node(name='cwtest', image=cw_img, size=cw_size) 
        
        #As the instance has no IP address yet, we must wait for the starting process to be completed
        #Periodacally read the list of existing Euca instance until it find a running instance with the CW instance name
        log.info("Waiting for the instance startup to be completed...")
        cw_node_started = None
        waited_time = 0
        while cw_node_started==None and waited_time<600:
            time.sleep(5)
            waited_time += 5
            nodes = conn.list_nodes()
            for n in nodes:
                if n.state == LC_NodeState.RUNNING and n.name == cw_node_launch.name:
                    cw_node_started = n
        
        if cw_node_started==None:
            log.error("Cannot retrieve the started instance of the CloudWorker")
            log.error("Destroying the instance previously created")
            cw_node_launch.destroy()
            return

        log.info("The instance is started as %s" % cw_node_started)

        self.specific_data = cw_node_started.id

        log.info("Connecting with SSH")
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.WarningPolicy())
        tries=0
        while tries<5:
            try:
                ssh.connect(cw_node_started.public_ip[0],username='root',key_filename='id_rsa')
                break
            except Exception as e:
                log.warn("Failing to connect")
                log.warn(e)                
                tries += 1
                if tries<5:
                    log.warn("Trying again in 5 seconds")
                    time.sleep(5)
                else:
                    log.error("Too much tries, unable to connect")
                    log.error("Destroying the CloudWorker instance")
                    cw_node_started.destroy()
                    return
        
        log.info("Executing commands")        
        for cmd in self.list_of_commands:
            log.info("Executing %s" % cmd)
            (stdin, stdout, stderr) =  ssh.exec_command(cmd)
            log.info("The output is \n%s" % "".join(stdout.readlines()) )
        
        log.info("All commands executed, closing connection")
        self.execution_is_success = True
        ssh.close()

class Euca_CW_Stopper(threading.Thread):
    """//Thread that stops a cloud worker in Eucalyptus//

    - Arguments:
     - instance_id -- The intance identifier of the cloud worker in Eucalyptus server
     
    - Attributes:
     - stop_is_success -- When the thread terminated, True if the cloud worker correctly stopped, False otherwise

    """     

    def __init__(self,instance_id):
        threading.Thread.__init__(self)
        if not isinstance(instance_id,str):
            raise Exception("The instance_id must be a string")
         
        self.euca_param = conf_func.read_config("euca.cfg")
        if not ("EUCA_ACCESS_KEY" in self.euca_param and \
                "EUCA_SECRET_KEY" in self.euca_param and \
                "EUCA_HOST" in self.euca_param and \
                "EUCA_CW_IMAGE_ID" in self.euca_param and \
                "EUCA_CW_SIZE_ID" in self.euca_param):
            raise Exception("Missing parameters in \"euca.cfg\"")
        
        if not "EUCA_PORT" in self.euca_param:
            self.euca_param["EUCA_PORT"] = 8773 

        self.instance_id = instance_id
        self.stop_is_success = False
        
    def run(self):
        log.info("Connecting to Eucalyptus at %s" % self.euca_param["EUCA_HOST"])
        
        Driver = LC_providers.get_driver(LC_ProviderType.EUCALYPTUS)
        conn = Driver(self.euca_param["EUCA_ACCESS_KEY"], self.euca_param["EUCA_SECRET_KEY"],\
                host=self.euca_param["EUCA_HOST"],port=self.euca_param["EUCA_PORT"]) 
        
        log.info("Retrieving the list of Eucalyptus instance")
        nodes = conn.list_nodes() 
        cw_node = None
        for n in nodes:
            if self.instance_id == n.id:
                cw_node=n

        if cw_node == None:
            log.error("Cannot find CloudWorker instance in Eucalyptus server")
            return 
        log.info("CloudWorker instance found as %s" % str(cw_node))

        r = cw_node.destroy()

        if r != True:
            log.error("Error when terminating the instance")
        else:
            log.info("Instance destroyed")
            self.stop_is_success=True





if __name__=="__main__":
    cw = Euca_CW_Starter(['ls','pwd'])
    cw.start()
    cw.join()
    if cw.execution_is_success:
        print "Press enter to destroy the cloud worker"
        x=raw_input()
        cwstop = Euca_CW_Stopper(cw.specific_data)
        cwstop.start()
