#! /usr/bin/python

"""//Program that monitor QoS-enabled batches//

- > This program manages batches previously created by the start_qos function by allocating cloud resources to batch being funded by users. It should be started periodically to perform monitoring operations. 

- File: monitor_batches.py

- Notes: 
 - A database configuration file named "db.cfg" must be present in the directory. 
 - A file named "creditsystem.cfg" containing a line "address=<url>", with <url> the url to access to the Credit System web server, must be present in the directory. 
 - A file named "oracle.cfg" containing a line "address=<url>", with <url> the url to access to the Oracle web server, must be present in the directory. 

- Algorithm:
``` 
For each of batches, this function checks 
  if the batch is not already computed by a CloudWorker, 
                        (!!Could change in the future!!)
   and if we at 90% of the batch completion
   and if the oracle-predicted benefit of using cloud 
                                              is enough, 
   and if the batch order has more credits than the 
                    credit cost predicted by the oracle, 
   and if there is one cloud worker available, 
      starting it and assigning the batch to it.
``` 

"""

import sys
import os
sys.path.append(os.path.join(sys.path[0],".."))
os.chdir(sys.path[0])
from tools import conf_func
from tools import web_func
from tools import db_func
from tools import log_func
log=log_func.get_spequlos_log(__file__)

import scheduler_func


CW_START_THRESHOLD = 0.9


def monitor_batches(conn):
    """Function implementing the batches monitoring
      
    Arguments:
    conn -- The scheduler's DB connection, as a MySQL.connections.Connection object

    """        
    if not isinstance(conn,db_func.db_connection) or conn.open==0:
        log.error("The conn argument must be an open connection to the database !")
        return None         
            
    c=conn.cursor()

    log.info("Retrieving QoSBatches list...")
    try:
        c.execute("SELECT batch_id,dg_name,batch_data,dg_type,dg_plugin_url FROM QoSBatches NATURAL JOIN DesktopGrid")
    except db_func.DBError as e:
        log.error("Error during the SQL command execution !")
        log.error(e)
        return None      
    res=c.fetchall()

    if len(res)==0:
        log.info("No QoSBatches to manage, exiting...")
    else:
        for r in res:
            try:
                batch_id=r[0]
                dg_name=r[1]
                batch_data=r[2]
                dg_type=r[3]
                dg_plugin_url=r[4]
            except ValueError as e:
                log.error("Error when retrieving information on batch %s" % str(r))
                log.error(e)
                return None

            log.info("Managing batch %s for %s@%s" % (batch_id,dg_type,dg_name))

            log.debug("Checking how many cloud workers are working on this batch")
            #This currently not used, 1 worker is started each time needed  

            try:
                c.execute("SELECT cloud_worker_id FROM CloudWorkers WHERE batch_id='%s'" % batch_id)
            except db_func.DBError as e:
                log.error("Error during the SQL command execution !")
                log.error(e)
                return None  
            res2=c.fetchall()

            n_workers = len(res2)
            log.info("%d Cloud workers are working on batch %s" % (n_workers,batch_id))

            if n_workers>=0:
                log.info("Checking batch completion...")            
                
                completion = scheduler_func.get_percent_completed(batch_id)
                if completion==None:
                    log.warn("No result from get_percent_completed, skipping this batch")
                elif completion<CW_START_THRESHOLD:
                    log.info("The batch completion is not advanced enough to launch cloud workers")
                elif completion>=1:
                    log.info("The batch completion is too advanced to launch cloud workers")
                    #TODO : Test this part
                    if n_workers==0:
                        log.info("No cloud workers are working on this batch, deregistering it...")
                        
                        log.info("Updating the scheduler database")
                        try:
                            c.execute("DELETE FROM QoSBatches WHERE batch_id='%s'" % batch_id)
                        except db_func.DBError as e:
                            log.error("Error during the SQL command execution !")
                            log.error(e)
                            return None      

                        log.info("Unconfigure grid %s@%s for QoS on batch %s" % \
                                    (dg_type,dg_name,batch_id))
                        scheduler_func.unconfigure_QoS(batch_id,dg_type,dg_plugin_url,batch_data)
                else:
                    log.info("The batch completion is higher than %s checking benefits of using cloud..." % CW_START_THRESHOLD)

                    (benef,cost) = calculate_cloud_benefit(batch_id)
                    if benef==None:
                        log.error("No result from calculate_cloud_benefit, skipping this batch")
                    elif benef<MIN_BENEFIT:
                        log.info("The benefit of using cloud is not sufficient, skipping this batch")
                    else:
                        log.info("This batch can benefit from the use of cloud  ! Checking if there is enough credits ordered...")

                        cred = get_credits(batch_id)
                        if cred==None:
                            log.error("No result from get_credit, skipping this batch")
                        elif cred<cost:
                            log.info("Not enough credits ordered to use cloud ressources, skipping this batch")
                        else:
                            log.info("There is enough credits to use cloud ressources !" +\
                                                    "Checking cloud worker availability...")                            

                            try:
                                c.execute("SELECT cloud_worker_id,cloud_tech,cloud_host,batch_id FROM CloudWorkers WHERE batch_id='None'")
                            except db_func.DBError as e:
                                log.error("Error during the SQL command execution !")
                                log.error(e)
                                return None  
                            res2=c.fetchall()

                            if len(res2)==0:
                                log.warning("No cloud worker available!! Cannot do anything for this batch now.")
                            else:
                                try:
                                    cw_id=int(res2[0][0])
                                except ValueError as e:
                                    log.error("'cloud_worker_id' in CloudWorkers table must be an integer, skipping this batch")
                                    log.error(e)
                                else:                                    
                                    cw_tech=res2[0][1]
                                    cw_host=res2[0][2]
                                    
                                    log.info("Cloud worker %d (%s@%s) is available, starting it" % (cw_id,cw_tech,cw_host))
                                    result_data = scheduler_func.start_cloud_worker(batch_id,dg_type,dg_name,batch_data,cw_id,cw_tech,cw_host)
                                    if result_data == None:
                                        log.error("CloudWorker starting failed !")
                                        return None

                                    log.info("Done ! Assigning it to %s and updating its data to %s" % (batch_id,result_data))
                                    try:
                                        c.execute("""UPDATE CloudWorkers SET batch_id='%s',cloud_worker_data='%s' 
                                            WHERE cloud_worker_id=%d""" % (batch_id,result_data,cw_id))
                                    except db_func.DBError as e:
                                        log.error("Error during the SQL command execution !")
                                        log.error(e)
                                        log.error("Stopping the CloudWorker !")
                                        scheduler_func.stop_cloud_worker(cw_id,cw_tech,cw_host,result_data)
                                        log.error("Stopping completed")
                                        return None  
 
                                    log.info("Done !")
    
                                    #TODO is this a bad hack ?
                                    if n_workers==0:
                                        log.info("Fist Cloudworker started! Configuring the grid to enable QoS for batch %s" % (batch_id))
                                        scheduler_func.configure_qos(batch_id,dg_type,dg_plugin_url,cw_id)
                                    
                                    #TODO ? First billing of the cloud worker use
                    log.info("End of managment for batch %s" % batch_id)


def get_credits(batch_id):
    """Remote call of Credit System's get_credits
    
    The function calls and retrieves the result of the get_credits function from the Credit System web server.
    It needs a file named "creditsystem.cfg" containing the line "address=<url>", with <url> the url to access to the QoS Info web server, to be present in the directory. 
    
    Arguments:
    batch_id -- The batch identifier, as a string
    
    The function returns the number of credits in the batch order, as an integer
    
    """  
    if not isinstance(batch_id,str):
        log.error("get_credits() batch_id argument must be a string")
        return None    
    
    params=conf_func.read_config("creditsystem.cfg")
    
    url="http://"+params['address']+"/get_credits.py?batch_id="+batch_id

    res=web_func.call(url)
                                                    
    if len(res)!=1:
        log.error("There should be 1 result from url %s" % url)
        return None
    try:
        r=int(res[0])
        if r<0:
            raise ValueError("Negative number of credit")
    except ValueError as e:
        log.error("get_credits result must be a positive integer")
        log.error(e)
        return None
    else:
        return r            



def calculate_cloud_benefit(batch_id):
    """Remote call of Oracle's calculate_cloud_benefit
    
    The function calls and retrieves the result of the calculate_cloud_benefit function from the Oracle web server.
    It needs a file named "oracle.cfg" containing the line "address=<url>", with <url> the url to access to the Oracle web server, to be present in the directory. 
    
    Arguments:
    batch_id -- The batch identifier, as a string
    
    The function returns (r,c), with r the ratio between the expected batch makespan (with tail problem) with batch makespan when using cloud ressources, as a float and c the exepected cost of cloud ressource usage, as an integer
    
    """ 
    if not isinstance(batch_id,str):
        log.error("calculate_cloud_benefit() batch_id argument must be a string")
        return None
             
    params=conf_func.read_config("oracle.cfg")
    
    url="http://"+params['address']+"/calculate_cloud_benefit.py?batch_id="+batch_id\
        +"%s" % "" if CURRENT_TIME is None else "current_time="+str(CURRENT_TIME)

    res=web_func.call(url)
                                                    
    if len(res)!=4:
        log.error("There should be 4 results from url %s" % url)
        return
    try:
        r1 = int(res[1])
        r2 = int(res[2])
        c = int(res[3])
        if r1<0:
            raise ValueError("2nd result is negative")
        if r2<0:
            raise ValueError("3rd result is negative")
        if c<0:
            raise ValueError("4th result is negative")
    except ValueError as e:
        log.error("calculate_cloud_benefit results must be positive integers")            
        log.error(e)            
        return None
    else:
        return float(r1)/r2,c

if __name__=="__main__":
    #Change this to specify the current time of the system (in Unix ticks) 
    CURRENT_TIME=None

    #To use a cloud worker, the cloud assisted makespan should be MIN_BENEFIT times lower than normal makespan (with tail problem)
    MIN_BENEFIT=1.5

    conn=db_func.connect(db_func.read_db_config("db.cfg"))
    monitor_batches(conn)
    db_func.disconnect(conn)



