#! /usr/bin/python

"""//Functions used by the scheduler monitor_batches and monitor_cloud_workers programs//

- File: scheduler_func.py

"""

import sys
import os
sys.path.append(os.path.join(sys.path[0],".."))
os.chdir(sys.path[0])
from tools import web_func
from tools import conf_func
from tools import log_func
log=log_func.get_spequlos_log(__file__)

def stop_cloud_worker(cw_id,cw_tech,cw_host,cw_data):
    """//Stop a cloud worker//

    - > According to the CloudWorker properties, such as the cloud technology it uses, this function load the appropriate implementation to stop the CloudWorker.
    
    - Arguments:
     - cw_id -- The CloudWorker identifier, as an integer (currently unused in this function)
     - cw_tech -- The cloud technology used by the CloudWorker, as a string. Currently, only "EC2", "EUCA", or "DummyCloudWorker" are implemented
     - cw_host -- The CloudWorker cloud host (currently unused)
     - cw_data -- Some data associated to the cloud worker, for technology-specific inforamtion (used to store the instance name for EC2 or EUCA)
    
    - The function returns True if the CloudWorker is correctly stopped, or False otherwise
    
    """     
    if cw_tech=="DummyCloudWorker":
        from dummy_cloudworker.dummy_cloudworker import Dummy_CW_Stopper as CW_Stopper
    elif cw_tech=="G5K":
        from g5k_cloudworker.g5k_cloudworker import G5K_CW_Stopper as CW_Stopper
    elif cw_tech=="EC2":    
        from libcloud_cloudworker.ec2_cloudworker import EC2_CW_Stopper as CW_Stopper
    elif cw_tech=="EUCA":    
        from libcloud_cloudworker.euca_cloudworker import Euca_CW_Stopper as CW_Stopper
    else:
        log.error("No implementation for %s cloud technology" % cw_tech)
        return
    
    cw = CW_Stopper(cw_data)
    cw.start()
    cw.join()
    if cw.stop_is_success:
        log.info("CloudWorker stopped")
        return True
    else:
        log.error("CloudWorker stop failed")
        return False

def start_cloud_worker(batch_id,dg_type,dg_name,batch_data,cw_id,cw_tech,cw_host):
    """//Start a cloud worker//

    - > According to the CloudWorker properties, such as the cloud technology it uses, this function load the appropriate implementation to start the CloudWorker. Then, according to the batch properties, such as the grid technologie it runs on, the function executes the appropriate commands on the CloudWorker to make it join the grid and participate to the batch computation
    
    - Arguments:
     - batch_id -- The batch identifier, as a string
     - dg_type -- The batch grid technology, as a string. Currently, only "XWHEP" or "BOINC" are implemented.
     - dg_name -- An unique name identifying the DG server. In the cloud worker, used to start the appropriate script (located to /root/spequlos/<dg_name>.sh)
     - batch_data -- Some data associated to the batch, for DG technology-specific information, as a string (used to store the group URI associated to the batch in XWHEP), as a string
     - cw_id -- The CloudWorker identifier, as an integer. Currently unused in this function.
     - cw_tech -- The cloud technology used by the CloudWorker, as a string. Currently, only "EC2", "EUCA", or "DummyCloudWorker" are implemented
     - cw_host -- The CloudWorker cloud host. Currently unused.
    
    - The function returns some cloud technology specific data if the CloudWorker is correctly started (for EC2 or EUCA, the instance name is returned)
    
    """
    if cw_tech=="DummyCloudWorker":
        from dummy_cloudworker.dummy_cloudworker import Dummy_CW_Starter as CW_Starter
    elif cw_tech=="G5K":
        from g5k_cloudworker.g5k_cloudworker import G5K_CW_Starter as CW_Starter
    elif cw_tech=="EC2":
        from libcloud_cloudworker.ec2_cloudworker import EC2_CW_Starter as CW_Starter
    elif cw_tech=="EUCA":
        from libcloud_cloudworker.euca_cloudworker import Euca_CW_Starter as CW_Starter
    else:
        log.error("No implementation for %s cloud technology" % cw_tech)
        return
   
    if dg_type=="XWHEP":
        list_of_commands = ['~/spequlos/%s.sh %s %s' % (dg_name,batch_id,batch_data) ]
    elif dg_type=="BOINC":
        list_of_commands = ['~/spequlos/%s.sh' % (dg_name) ]
    else: 
        log.error("No implementation for %s grid technology" % dg_type)
        return
    
    cw = CW_Starter(list_of_commands)
    cw.start()
    cw.join()
    if cw.execution_is_success:
        log.info("CloudWorker is started")
        return cw.specific_data
    else:
        log.error("CloudWorker starting failed")
        return None


def preconfigure_QoS(batch_id,dg_type,dg_plugin_url):
    """//Preconfigure a Desktop Grid to be able to handle QoS//

    - > The preconfiguration is a configuration step that must be performed in the DG before the batch submission, i.e. when the user calls the start_qos function

    - Arguments:
     - batch_id -- The batch identifier, as a string
     - dg_type -- The grid type used for the batch computation, as a string. Currently, only XWHEP needs some preconfiguration
     - dg_plugin_url -- The URL to the DG plugins, as a string.
    
    - The function returns some datas, as a string, specific to the DG technology, as output of the preconfiguration (for XWHEP its the group URI associated with the batch_id) or True if there is no specific data or None is the preconfiguration failed
    
    """     
    if not isinstance(batch_id,str):
        log.error("The batch_id argument must be a string")
        return None
    if not isinstance(dg_type,str):
        log.error("The dg_type argument must be a string")
        return None
    if not isinstance(dg_plugin_url,str):
        log.error("The dg_plugin_url argument must be a string")
        return None
 
    log.info("Checking the DG technology used")
    if dg_type=="XWHEP":    
        log.info("For XWHEP, we have to create a job group associated to our batch. Calling corresponding plugin...")

        if dg_plugin_url[-1:] == "/":
            log.debug("Removing trailing slash from dg_plugin_url")
            dg_plugin_url=dg_plugin_url[0:-1]

        url=dg_plugin_url+"/XWHEP-addgroup.php?batch_id="+batch_id
        res=web_func.call(url)
        
        if len(res)!=1:
            log.error("There should be 1 results from url %s" % url)
            return None
        if res[0]=="False":
            log.error("The %s server at %s returned an error with batch_id %s" % (dg_type,dg_plugin_url,batch_id))
            return None
        else:
            uri = res[0]
            log.info("The %s server at %s returned the URI %s for batch_id %s" % (dg_type,dg_plugin_url,uri,batch_id))
            return uri
    else:
        log.info("No preconfiguration needed for %s grid technology" % dg_type)
        return True




def configure_qos(batch_id,dg_type,dg_plugin_url,cw_id):
    """//Configure a Desktop Grid to start QoS//

    - > The configuration is a step performed once the requierment to start QoS are filled (ie. credit ordered, oracle agreement, cloudworker started, etc.)

    - Arguments:
     - batch_id -- The batch identifier, as a string
     - dg_type -- The grid technology used for the batch computation, as a string. Currently, only BOINC needs some configuration
     - dg_plugin_url -- The URL to the DG plugins, as a string.
     - cw_id -- The cloud worker identifier, as an interger. Currently unused.

    - The function returns True if the configuration succeed or None otherwise
    
    """     
    if not isinstance(batch_id,str):
        log.error("The batch_id argument must be a string")
        return None
    if not isinstance(dg_type,str):
        log.error("The dg_type argument must be a string")
        return None
    if not isinstance(dg_plugin_url,str):
        log.error("The dg_plugin_url argument must be a string")
        return None
    if not isinstance(cw_id,int):
        log.error("The cw_id argument must be an integer")
        return None
 
    log.info("Checking the DG technology used")
    if dg_type=="BOINC":    
        log.info("For BOINC, we have to ask for association between uncompleted workunit and cloud worker. Calling corresponding plugin...")

        if dg_plugin_url[-1:] == "/":
            log.debug("Removing trailing slash from dg_plugin_url")
            dg_plugin_url=dg_plugin_url[0:-1]

        url=dg_plugin_url+"/assign_batch_to_cloud?batch="+batch_id
        res=web_func.call(url)
        
        if len(res)!=1:
            log.error("There should be 1 results from url %s" % url)
            return None
        if res[0]!="True":
            log.error("The %s server at %s returned an error with batch_id %s" % (dg_type,dg_plugin_url,batch_id))
            return None
        else:
            log.info("Configuration succeed !")
            return True
    else:
        log.info("No preconfiguration needed for %s grid technology" % dg_type)
        return True


def unconfigure_QoS(batch_id,dg_type,dg_plugin_url,batch_data,cw_id=-1,cw_data=''):
    """//Unconfigure a Desktop Grid when QoS operation finished//

    - > The unconfiguration is a configuration step that must be performed in the DG when the QoS operations related to a batch is finnished, i.e. when the batch is completed or when no more resources are available to support QoS

    - Arguments:
     - batch_id -- The batch identifier, as a string
     - dg_type -- The grid technology used for the batch computation, as a string. Currently, only XWHEP needs some unconfiguration
     - dg_plugin_url -- The URL to the DG plugins, as a string.
     - batch_data -- Some data associated to the batch, for DG technology-specific information, as a string (used to store the group URI associated to the batch in XWHEP), as a string
     - cw_id -- The CloudWorker identifier, as an integer. Currently unused in this function.
     - cw_data -- Some data associated to the cloud worker, for technology-specific information, as a string (used to store the instance name for EC2), as a string. Currently unused in this function.

    - The function returns True if the unconfiguration succeed or None otherwise
    
    """     
    if not isinstance(batch_id,str):
        log.error("The batch_id argument must be a string")
        return None
    if not isinstance(dg_type,str):
        log.error("The dg_type argument must be a string")
        return None
    if not isinstance(dg_plugin_url,str):
        log.error("The dg_plugin_url argument must be a string")
        return None
    if not isinstance(batch_data,str):
        log.eror("The batch_data argument must be a string")
        return None
    if not isinstance(cw_id,int):
        log.error("The cw_id argument must be an integer")
        return None
    if not isinstance(cw_data,str):
        log.error("The cw_data argument must be a string")
        return None

    log.info("Checking which DG technology is used")
    if dg_type=="XWHEP":  
        
        if dg_plugin_url[-1:] == "/":
            log.debug("Removing trailing slash from dg_plugin_url")
            dg_plugin_url=dg_plugin_url[0:-1]

        url=dg_plugin_url+"/XWHEP-rmgroup.php?batch_id_uri="+batch_data
        log.info("For XWHEP, we have to remove the job group associated to our batch. Calling corresponding plugin at url %s" % url)

        res=web_func.call(url)
        if len(res)!=1:
            log.error("There should be 1 results from url %s" % url)
            return None
        if res[0]!="True":
            log.error("The %s server at %s returned an error with uri %s" % (dg_type,dg_plugin_url,batch_data))
            return None
        else:
            log.info("The %s server at %s deleted the URI %s associated to batch_id %s" % (dg_type,dg_plugin_url,batch_data,batch_id))
            return True 
    else:
        log.info("No preconfiguration needed for %s grid technology" % dg_type)
        return True



def get_percent_completed(batch_id):
    """//Remote call of Oracle's get_percent_completed//
    
    The function calls and retrieves the result of the get_percent_completed function from the Oracle web server.
    It needs a file named "oracle.cfg" containing the line "address=<url>", with <url> the url to access to the Oracle web server, to be present in the directory. 
    
    Arguments:
    batch_id -- The batch identifier, as a string
    
    The function returns the batch completion ratio, as a float
    
    """ 
    if not isinstance(batch_id,str):
        log.error("calculate_cloud_benefit() batch_id argument must be a string")
        return None

    #Change this to specify the current time of the system (in Unix ticks) 
    CURRENT_TIME=None             

    params=conf_func.read_config("oracle.cfg")
    
    url="http://"+params['address']+"/get_percent_completed.py?batch_id="+batch_id\
        +"%s" % "" if CURRENT_TIME is None else "current_time="+str(CURRENT_TIME)

    res=web_func.call(url)
                                                    
    if len(res)!=1:
        log.error("There should be 1 results from url %s" % url)
        return
    try:
        if res[0] == "None":
            log.error("Oracle get_percent_completed did not returned result")            
            return None        
        p = float(res[0])
        if p<0:
            raise ValueError("completion is negative")
        if p>1:
            raise ValueError("completion is greater than 1")
    except ValueError as e:
        log.error("get_percent_completed result must be a float between 0 and 1")            
        log.error(e)            
        return None
    else:
        return p


def is_compute_finished(cw_id,cw_tech,cw_host,batch_id,dg_type,dg_name):
    if get_percent_completed(batch_id)==1:
        return True
    else:
        return False
